/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wPeriodic.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <list>
#include <algorithm>
using namespace heat;

HEAT_API std::ostream &
operator<<(std::ostream &out, NodePair const &obj)
{
    out << "node #" << obj.n->no << " <=> node #" << obj.d->no;
    return out;
}

NodePair::NodePair(Node *_n, Node *_d, double tol) : n(_n), d(_d)
{
    Eigen::Vector3d l = d->pos - n->pos;
    if (fabs(l(0)) > tol && fabs(l(1)) > tol)
    {
        std::stringstream s;
        s << " node #" << n->no << " and node #" << d->no << " are not facing each other!\n";
        throw std::runtime_error(s.str());
    }
}

Periodic::Periodic(std::shared_ptr<heat::Problem> _pbl,
                   double _TM, double _dTxM, double _dTyM) : wSharedObject(), pbl(_pbl)
{
    // pbl->pdic = this;
    reset(_TM, _dTxM, _dTyM);
    verbose = false;
}

void Periodic::reset(double _TM, double _dTxM, double _dTyM)
{
    TM = _TM;
    dTxM = _dTxM;
    dTyM = _dTyM;
    qM = Eigen::Vector3d::Zero();
}

void Periodic::write(std::ostream &out) const
{
    out << "heat::Periodic\n";
    out << "\tT_M     = " << TM << '\n';
    out << "\tgradT_M = (" << dTxM << ", " << dTyM << ")\n";
}

void Periodic::init()
{
    if (!LR.empty())
        return;

    // std::cout << "Periodic::init()\n";

    // build a list of nodes related to all the media
    if (verbose)
    {
        std::cout << "extracting nodes from " << pbl->media.size()
                  << " medi" << ((pbl->media.size() == 1) ? "um" : "a") << "...\n";
    }
    std::list<Node *> nlist;
    for (auto mat : pbl->media)
        for (auto e : mat->tag->elems)
            for (auto n : e->nodes)
                nlist.push_back(n);
    nlist.sort();
    nlist.unique();
    if (verbose)
        std::cout << "number of nodes:" << nlist.size() << '\n';
    if (nlist.empty())
        throw std::runtime_error("Periodic::init() - no mesh!");

    // compute bbox (p1,p3)

    Eigen::Vector3d p1 = (*nlist.begin())->pos;
    Eigen::Vector3d p3 = p1;
    for (auto n : nlist)
    {
        Eigen::Vector3d p = n->pos;
        if (p(0) < p1(0))
            p1(0) = p(0);
        if (p(0) > p3(0))
            p3(0) = p(0);
        if (p(1) < p1(1))
            p1(1) = p(1);
        if (p(1) > p3(1))
            p3(1) = p(1);
    }
    if (verbose)
    {
        std::cout << "bounding box:\n";
        std::cout << "\tp1=" << p1 << '\n';
        std::cout << "\tp3=" << p3 << '\n';
    }

    // compute smallest distance between 2 nodes (mesh characteristic length)
    double lc = std::numeric_limits<double>::max();
    for (auto mat : pbl->media)
        for (auto e : mat->tag->elems)
            for (auto n1 : e->nodes)
                for (auto n2 : e->nodes)
                    if (n1 != n2)
                    {
                        double l = (n1->pos - n2->pos).norm();
                        if (l < lc)
                            lc = l;
                    }
    if (verbose)
        std::cout << "characteristic length = " << lc << '\n';

    // build ordered lists of nodes on boundaries
    auto nL = nodesOnEdge(nlist, lc, p1, 0, "L");
    auto nR = nodesOnEdge(nlist, lc, p3, 0, "R");
    auto nB = nodesOnEdge(nlist, lc, p1, 1, "B");
    auto nT = nodesOnEdge(nlist, lc, p3, 1, "T");

    // corners detection
    double eps = lc / 10.;

    corners.resize(4);

    {
        if (verbose)
            std::cout << "detecting corners...\n";
        auto itP1 = nL.begin();
        auto itP2 = nR.begin();
        auto itP4 = --nL.end();
        auto itP3 = --nR.end();

        corners[0] = *itP1;
        corners[1] = *itP2;
        corners[2] = *itP3;
        corners[3] = *itP4;

        NodePair(*itP1, *itP2, eps); // error check
        NodePair(*itP2, *itP3, eps); // error check
        NodePair(*itP3, *itP4, eps); // error check
        NodePair(*itP4, *itP1, eps); // error check

        if (verbose)
            for (auto n : corners)
                std::cout << '\t' << *n << '\n';

        // build LR node pairs
        if (verbose)
            std::cout << "detecting node pairs : left-right (LR)...\n";
        for (auto itL = ++itP1, itR = ++itP2; itL != itP4 && itR != itP3; ++itL, ++itR)
            LR.push_back(NodePair(*itL, *itR, eps));
        /*
        for(auto np : LR)
            std::cout << '\t' << np << '\n';
        */
    }

    // build TB node pairs
    {
        auto itP1 = nB.begin();
        auto itP2 = --nB.end();
        auto itP4 = nT.begin();
        auto itP3 = --nT.end();

        if (verbose)
            std::cout << "detecting node pairs : bottom-top (BT)...\n";
        for (auto itB = ++itP1, itT = ++itP4; itB != itP2 && itT != itP3; ++itB, ++itT)
            BT.push_back(NodePair(*itB, *itT, eps));
        /*
        for(auto np : BT)
            std::cout << '\t' << np << '\n';
            */
    }
}

std::vector<Node *>
Periodic::nodesOnEdge(std::list<Node *> const &nlist,
                      double lc, Eigen::Vector3d const &pt,
                      int c, std::string const &name) const
{
    struct LessThan
    {
        int comp;
        LessThan(int c) : comp(c) {}
        inline bool operator()(Node *const &o1, Node *const &o2) const
        {
            return (o1->pos(comp) < o2->pos(comp));
        }
    };

    std::vector<Node *> lnods;
    double eps = lc / 10.;
    double xF = pt(c);
    for (auto n : nlist)
    {
        double px = n->pos(c);
        if (xF - eps < px && px < xF + eps)
            lnods.push_back(n);
    }
    std::sort(lnods.begin(), lnods.end(), LessThan((c == 1) ? 0 : 1));

    if (verbose)
        std::cout << "nodes on " << name << " boundary: " << lnods.size() << '\n';
    if (lnods.size() < 2)
    {
        std::stringstream str;
        str << "Periodic::nodesOnEdge - not enough nodes on " << name << " boundary!\n";
        throw std::runtime_error(str.str());
    }

    /*
    for(auto n : lnods)
        std::cout << '\t' << *n << '\n';
    */
    return lnods;
}
